use rocket::response::status::BadRequest;
use serde_json::json;
use std::str::FromStr;

/// API error response with a JSON payload.
pub type ApiError = BadRequest<serde_json::Value>;

/// Fallible API response.
///
/// This type represents the result of an API operation that can either succeed with a
/// JSON payload (`serde_json::Value`) or fail with an `ApiError`.
pub type FallibleApiResponse = Result<serde_json::Value, ApiError>;

/// A trait to generate a Model API representation in JSON format.
///
/// This trait is intended for data models that need to be serialized
/// into a JSON representation suitable for public APIs.
pub trait ToAPI: Sized {
    /// Generate a JSON representation of the model for public API use.
    fn api(&self) -> impl std::future::Future<Output = serde_json::Value>;
}

/// Converts a slice of items implementing the `ToAPI` trait into a `Vec` of JSON values.
///
/// This function is asynchronous and iterates over a slice of items, calling the `api`
/// method on each to generate its JSON representation. The results are collected into
/// a `Vec<serde_json::Value>`.
///
/// # Arguments
/// * `items` - A slice of items implementing the `ToAPI` trait.
///
/// # Returns
/// An asynchronous computation that resolves to a `Vec` of JSON values.
pub async fn vec_to_api(items: &[impl ToAPI]) -> Vec<serde_json::Value> {
    let mut ret = Vec::with_capacity(items.len());

    for e in items {
        ret.push(e.api().await);
    }

    ret
}

/// Converts a string into a `Uuid`.
///
/// This function attempts to parse a string as a UUID. If the parsing fails,
/// it returns an `ApiError` indicating the failure.
///
/// # Arguments
/// * `id` - A string slice representing the UUID to be parsed.
///
/// # Returns
/// * `Ok(uuid::Uuid)` if the string is a valid UUID.
/// * `Err(ApiError)` if the string is not a valid UUID, containing a descriptive error message.
pub fn to_uuid(id: &str) -> Result<uuid::Uuid, ApiError> {
    uuid::Uuid::from_str(id).map_err(|_| no_uuid_error())
}

/// Generates an `ApiError` indicating that the input is not a valid UUID.
///
/// This function is used internally to standardize error responses for invalid UUIDs.
///
/// # Returns
/// * `ApiError` - A `BadRequest` error with a JSON payload describing the issue.
#[must_use]
pub fn no_uuid_error() -> ApiError {
    api_error("No valid UUID")
}

/// Generates a custom `ApiError` with a given message.
///
/// This function creates a `BadRequest` response with a JSON payload containing
/// an error message, making it useful for consistent error reporting.
///
/// # Arguments
/// * `msg` - A string slice containing the error message to be included in the response.
///
/// # Returns
/// * `ApiError` - A `BadRequest` error with a JSON payload describing the issue.
#[must_use]
pub fn api_error(msg: &str) -> ApiError {
    BadRequest(json!({
        "error": msg
    }))
}

/// A `Pager` that manages paginated items, with the ability to handle incomplete data.
pub struct Pager<T> {
    inner: Vec<T>,
    pub items_per_page: u64,
    complete_at: Option<u64>,
}

impl<T> Pager<T> {
    /// Creates a new `Pager` instance with the given items and items per page.
    ///
    /// # Arguments
    /// * `items` - A vector of items to paginate.
    /// * `per_page` - Number of items per page.
    ///
    /// # Returns
    /// A new `Pager` instance.
    #[must_use]
    pub const fn new(items: Vec<T>, per_page: u64) -> Self {
        Self {
            inner: items,
            items_per_page: per_page,
            complete_at: None,
        }
    }

    /// Creates a new `Pager` instance for an incomplete dataset, starting from a specific page.
    ///
    /// # Arguments
    /// * `items` - A vector of items to paginate.
    /// * `per_page` - Number of items per page.
    /// * `at_page` - The page where data starts, meaning there are no pages before it.
    ///
    /// # Returns
    /// A new `Pager` instance.
    #[must_use]
    pub const fn new_incomplete(items: Vec<T>, per_page: u64, at_page: u64) -> Self {
        Self {
            inner: items,
            items_per_page: per_page,
            complete_at: Some(at_page),
        }
    }

    /// Calculates the offset for the given page number.
    ///
    /// This method considers whether the pager has an incomplete dataset and adjusts the offset accordingly.
    ///
    /// # Arguments
    /// * `page` - The page number for which the offset is calculated.
    ///
    /// # Returns
    /// The calculated offset.
    #[must_use]
    pub const fn offset(&self, page: u64) -> u64 {
        if let Some(incomplete) = self.complete_at {
            let page = page - incomplete;
            return self.items_per_page * page;
        }

        self.items_per_page * (page - 1)
    }

    /// Retrieves the items for a specific page.
    ///
    /// # Arguments
    /// * `page` - The page number to retrieve.
    ///
    /// # Panics
    /// Panics if trying to access a page before the `complete_at` page in an incomplete pager.
    ///
    /// # Returns
    /// A vector of items on the requested page.
    #[must_use]
    pub fn page(&self, page: u64) -> Vec<&T> {
        if let Some(incomplete) = self.complete_at {
            assert!(
                page >= incomplete,
                "Tried to access illegal page on incomplete Pager"
            );
        }

        self.inner
            .iter()
            .skip(self.offset(page).try_into().unwrap())
            .take(self.items_per_page.try_into().unwrap())
            .collect()
    }
}

/// A `GeneratedPager` is a paginated generator that fetches items dynamically from a generator function.
pub struct GeneratedPager<T, G, I>
where
    G: Fn(I, u64, u64) -> futures::future::BoxFuture<'static, Vec<T>>,
{
    generator: G,
    pub items_per_page: u64,
    _marker: std::marker::PhantomData<(T, I)>,
}

impl<T, G, I> GeneratedPager<T, G, I>
where
    G: Fn(I, u64, u64) -> futures::future::BoxFuture<'static, Vec<T>>,
{
    /// Creates a new `GeneratedPager` instance with the provided generator and pagination settings.
    ///
    /// # Arguments
    /// * `generator` - A function that generates a page of items based on the input and pagination settings.
    /// * `items_per_page` - Number of items per page.
    ///
    /// # Generator
    /// The generator function should take the following arguments:
    /// * `input` - Generic input to the generator
    /// * `offset` - Offset value
    /// * `limit` - Limit value (items per page)
    ///
    /// # Returns
    /// A new `GeneratedPager` instance.
    pub const fn new(generator: G, items_per_page: u64) -> Self {
        Self {
            generator,
            items_per_page,
            _marker: std::marker::PhantomData,
        }
    }

    /// Calculates the offset for the given page number.
    ///
    /// # Arguments
    /// * `page` - The page number for which the offset is calculated.
    ///
    /// # Returns
    /// The calculated offset.
    pub const fn offset(&self, page: u64) -> u64 {
        self.items_per_page * (page - 1)
    }

    /// Asynchronously retrieves the items for a specific page from the generator.
    ///
    /// # Arguments
    /// * `page` - The page number to retrieve.
    /// * `input` - The input that is passed to the generator.
    ///
    /// # Returns
    /// A vector of items on the requested page.
    pub async fn page(&self, page: u64, input: I) -> Vec<T> {
        let offset = self.offset(page);
        (self.generator)(input, offset, self.items_per_page).await
    }

    /// Converts the `GeneratedPager` into a regular `Pager` for a given page of items.
    ///
    /// This method allows you to use a `GeneratedPager` like a regular `Pager`, with a pre-generated dataset.
    ///
    /// # Arguments
    /// * `page` - The page number to retrieve.
    /// * `input` - The input to pass to the generator function.
    ///
    /// # Returns
    /// A `Pager` instance containing the requested page of items.
    pub async fn pager(&self, page: u64, input: I) -> Pager<T> {
        let content = self.page(page, input).await;
        Pager::new_incomplete(content, self.items_per_page, page)
    }
}
