use rocket::Request;
use rocket::Response;
use rocket::http::Header;
use rocket::http::Status;
use rocket::response::Responder;
use std::io::Cursor;

// TODO: Implement file based response

pub struct DataResponse {
    data: Vec<u8>,
    content_type: String,
    cache_duration: Option<u64>,
}

impl DataResponse {
    #[must_use]
    pub const fn new(data: Vec<u8>, content_type: String, cache_duration: Option<u64>) -> Self {
        Self {
            data,
            content_type,
            cache_duration,
        }
    }
}

#[rocket::async_trait]
impl<'r> Responder<'r, 'static> for DataResponse {
    fn respond_to(self, req: &'r Request<'_>) -> rocket::response::Result<'static> {
        // Handle Range requests
        if let Some(range) = req.headers().get_one("Range") {
            if let Some((start, end)) = parse_range_header(range, self.data.len()) {
                // TODO : Reject invalid ranges
                // TODO : Multiple ranges?

                let sliced_data = &self.data[start..=end];
                return Ok(Response::build()
                    .header(Header::new(
                        "Content-Range",
                        format!("bytes {}-{}/{}", start, end, self.data.len()),
                    ))
                    .header(Header::new("Accept-Ranges", "bytes"))
                    .header(Header::new("Content-Type", self.content_type.clone()))
                    .status(Status::PartialContent)
                    .streamed_body(Cursor::new(sliced_data.to_vec()))
                    .finalize());
            }
        }

        // Add caching headers for static files
        let cache_control_header = self.cache_duration.map_or_else(
            || Header::new("Cache-Control", "no-cache"),
            |duration| Header::new("Cache-Control", format!("public, max-age={duration}")),
        );

        Ok(Response::build()
            .header(cache_control_header)
            .header(Header::new("Accept-Ranges", "bytes"))
            .header(Header::new("Content-Type", self.content_type))
            .streamed_body(Cursor::new(self.data))
            .finalize())
    }
}

/// Parse the Range header and return the start and end indices.
fn parse_range_header(range: &str, total_len: usize) -> Option<(usize, usize)> {
    if !range.starts_with("bytes=") {
        return None;
    }

    let range = &range[6..];
    let parts: Vec<&str> = range.split('-').collect();

    if parts.len() != 2 {
        return None;
    }

    let start = parts[0].parse::<usize>().ok();
    let end = parts[1].parse::<usize>().ok();

    match (start, end) {
        (Some(start), Some(end)) if start <= end && end < total_len => Some((start, end)),
        (Some(start), None) if start < total_len => Some((start, total_len - 1)),
        (None, Some(end)) if end < total_len => Some((0, end)),
        _ => None,
    }
}
