use maud::{PreEscaped, html};

use crate::{
    request::{RequestContext, StringResponse},
    ui::UIWidget,
};

/// Represents the HTML structure of a page shell, including the head, body class, and body content.
///
/// This structure is used to construct the overall HTML structure of a page, making it easier to generate consistent HTML pages dynamically.
pub struct Shell {
    /// The HTML content for the `<head>` section of the page.
    head: PreEscaped<String>,
    /// An optional class attribute for the `<body>` element.
    body_class: String,
    /// The HTML content for the static body portion.
    body_content: PreEscaped<String>,
    ui: bool,
}

impl Shell {
    /// Constructs a new `Shell` instance with the given head content, body content, and body class.
    ///
    /// # Arguments
    /// * `head` - The HTML content for the page's head.
    /// * `body_content` - The HTML content for the body of the page.
    /// * `body_class` - An optional class to apply to the `<body>` element.
    ///
    /// # Returns
    /// A `Shell` instance encapsulating the provided HTML content and attributes.
    #[must_use]
    pub fn new<T: UIWidget + 'static, C: UIWidget + 'static, B: UIWidget + 'static>(
        head: T,
        body_content: B,
        body_class: C,
    ) -> Self {
        Self {
            head: head.render(),
            body_class: body_class.extended_class().join(" "),
            body_content: body_content.render(),
            ui: false,
        }
    }

    pub fn use_ui(mut self) -> Self {
        self.ui = true;
        self
    }

    /// Renders the full HTML page using the shell structure, with additional content and a title.
    ///
    /// # Arguments
    /// * `content` - The additional HTML content to render inside the main content div.
    /// * `title` - The title of the page, rendered inside the `<title>` element.
    ///
    /// # Returns
    /// A `PreEscaped<String>` containing the full HTML page content.
    #[must_use]
    pub fn render(&self, content: PreEscaped<String>, title: &str) -> PreEscaped<String> {
        html! {
                html {
                    head {
                        title { (title) };
                        @if self.ui {
                            script src="https://cdn.tailwindcss.com" {};
                            script src="/assets/htmx.min.js" {};
                            script src="/assets/flowbite.min.js" {};
                            link href="/assets/flowbite.min.css" rel="stylesheet" {};
                            meta name="viewport" content="width=device-width, initial-scale=1.0";
                        };
                        (self.head)
                    };
                    @if !self.body_class.is_empty() {
                    body class=(self.body_class) {
                        (self.body_content);

                        div id="main_content" {
                            (content)
                        };
                    };
                } @else {
                    body {
                        (self.body_content);

                        div id="main_content" {
                            (content)
                        };
                    };
                }
            }
        }
    }

    /// Renders a full page or an HTMX-compatible fragment based on the request context.
    ///
    /// If the request is not an HTMX request, this function uses the provided shell to generate
    /// a full HTML page. If it is an HTMX request, only the provided content is rendered.
    ///
    /// # Arguments
    /// * `content` - The HTML content to render.
    /// * `title` - The title of the page for full-page rendering.
    /// * `ctx` - The `RequestContext` containing request metadata.
    ///
    /// # Returns
    /// A `StringResponse`
    pub async fn render_page(
        &self,
        content: PreEscaped<String>,
        title: &str,
        ctx: RequestContext,
    ) -> StringResponse {
        if ctx.is_htmx {
            (
                rocket::http::Status::Ok,
                (rocket::http::ContentType::HTML, content.into_string()),
            )
        } else {
            (
                rocket::http::Status::Ok,
                (
                    rocket::http::ContentType::HTML,
                    self.render(content, title).into_string(),
                ),
            )
        }
    }
}
