use maud::{Markup, Render, html};

use crate::ui::{UIWidget, color::UIColor};

pub enum BorderSize {
    _0,
    _2,
    _4,
    _8,
}

impl BorderSize {
    pub const fn to_value(&self) -> &str {
        match self {
            BorderSize::_0 => "0",
            BorderSize::_2 => "2",
            BorderSize::_4 => "4",
            BorderSize::_8 => "8",
        }
    }
}

pub enum BorderSide {
    X,
    Y,
    Start,
    End,
    Top,
    Right,
    Bottom,
    Left,
}

impl BorderSide {
    pub const fn to_value(&self) -> &str {
        match self {
            BorderSide::X => "x",
            BorderSide::Y => "y",
            BorderSide::Start => "s",
            BorderSide::End => "e",
            BorderSide::Top => "t",
            BorderSide::Right => "r",
            BorderSide::Bottom => "b",
            BorderSide::Left => "l",
        }
    }
}

pub enum BorderStyle {
    Solid,
    Dashed,
    Dotted,
    Double,
    Hidden,
    None,
}

impl BorderStyle {
    pub const fn to_value(&self) -> &str {
        match self {
            BorderStyle::Solid => "border-solid",
            BorderStyle::Dashed => "border-dashed",
            BorderStyle::Dotted => "border-dotted",
            BorderStyle::Double => "border-double",
            BorderStyle::Hidden => "border-hidden",
            BorderStyle::None => "border-none",
        }
    }
}

#[allow(non_snake_case)]
pub fn Border<T: UIWidget + 'static>(inner: T) -> BorderWidget {
    BorderWidget(Box::new(inner), None, None, None, None)
}

pub struct BorderWidget(
    Box<dyn UIWidget>,
    Option<BorderSize>,
    Option<BorderSide>,
    Option<Box<dyn UIColor>>,
    Option<BorderStyle>,
);

impl BorderWidget {
    #[must_use]
    pub fn size(mut self, size: BorderSize) -> Self {
        self.1 = Some(size);
        self
    }

    #[must_use]
    pub const fn side(mut self, side: BorderSide) -> Self {
        self.2 = Some(side);
        self
    }

    #[must_use]
    pub const fn style(mut self, style: BorderStyle) -> Self {
        self.4 = Some(style);
        self
    }

    #[must_use]
    pub fn color<C: UIColor + 'static>(mut self, color: C) -> Self {
        self.3 = Some(Box::new(color));
        self
    }

    fn border_class(&self) -> String {
        if let Some(side) = &self.2 {
            if let Some(size) = &self.1 {
                return format!("border-{}-{}", side.to_value(), size.to_value());
            }
        } else if let Some(size) = &self.1 {
            return format!("border-{}", size.to_value());
        }

        "border".to_owned()
    }
}

impl Render for BorderWidget {
    fn render(&self) -> Markup {
        self.render_with_class("")
    }
}

impl UIWidget for BorderWidget {
    fn can_inherit(&self) -> bool {
        true
    }

    fn base_class(&self) -> Vec<String> {
        let mut ret = vec![self.border_class()];

        if let Some(color) = &self.3 {
            ret.push(format!("border-{}", color.color_class()));
        }

        if let Some(style) = &self.4 {
            ret.push(style.to_value().to_string());
        }

        ret
    }

    fn extended_class(&self) -> Vec<String> {
        let mut c = self.base_class();
        c.extend_from_slice(&self.0.extended_class());
        c
    }

    fn render_with_class(&self, class: &str) -> Markup {
        if self.0.as_ref().can_inherit() {
            self.0
                .as_ref()
                .render_with_class(&format!("{} {class}", self.base_class().join(" ")))
        } else {
            html! {
                div class=(format!("{} {class}", self.base_class().join(" "))) {
                    (self.0.as_ref())
                }
            }
        }
    }
}

pub enum OutlineStyle {
    Solid,
    Dashed,
    Dotted,
    Double,
    None,
}

impl OutlineStyle {
    pub const fn to_value(&self) -> &str {
        match self {
            OutlineStyle::Solid => "outline",
            OutlineStyle::Dashed => "outline-dashed",
            OutlineStyle::Dotted => "outline-dotted",
            OutlineStyle::Double => "outline-double",
            OutlineStyle::None => "outline-none",
        }
    }
}

#[allow(non_snake_case)]
pub fn Outline<T: UIWidget + 'static>(width: u32, inner: T) -> OutlineWidget {
    OutlineWidget(Box::new(inner), width, None, None, 0)
}

pub struct OutlineWidget(
    Box<dyn UIWidget>,
    u32,
    Option<Box<dyn UIColor>>,
    Option<OutlineStyle>,
    u32,
);

impl OutlineWidget {
    #[must_use]
    pub const fn offset(mut self, offset: u32) -> Self {
        self.4 = offset;
        self
    }

    #[must_use]
    pub const fn style(mut self, style: OutlineStyle) -> Self {
        self.3 = Some(style);
        self
    }

    #[must_use]
    pub fn color<C: UIColor + 'static>(mut self, color: C) -> Self {
        self.2 = Some(Box::new(color));
        self
    }
}

impl Render for OutlineWidget {
    fn render(&self) -> Markup {
        self.render_with_class("")
    }
}

impl UIWidget for OutlineWidget {
    fn can_inherit(&self) -> bool {
        true
    }

    fn base_class(&self) -> Vec<String> {
        let class = match self.1 {
            0 => "outline-0",
            1 => "outline-1",
            2 => "outline-2",
            4 => "outline-4",
            8 => "outline-8",
            _ => &format!("outline-[{}px]", self.1),
        };

        let mut ret = vec![class.to_string()];

        if let Some(color) = &self.2 {
            ret.push(format!("outline-{}", color.color_class()));
        }

        if let Some(style) = &self.3 {
            ret.push(style.to_value().to_string());
        }

        ret.push(match self.4 {
            0 => "outline-offset-0".to_string(),
            1 => "outline-offset-1".to_string(),
            2 => "outline-offset-2".to_string(),
            4 => "outline-offset-4".to_string(),
            8 => "outline-offset-8".to_string(),
            _ => format!("outline-offset-[{}px]", self.4),
        });

        ret
    }

    fn extended_class(&self) -> Vec<String> {
        let mut c = self.base_class();
        c.extend_from_slice(&self.0.extended_class());
        c
    }

    fn render_with_class(&self, class: &str) -> Markup {
        if self.0.as_ref().can_inherit() {
            self.0
                .as_ref()
                .render_with_class(&format!("{} {class}", self.base_class().join(" ")))
        } else {
            html! {
                div class=(format!("{} {class}", self.base_class().join(" "))) {
                    (self.0.as_ref())
                }
            }
        }
    }
}

#[allow(non_snake_case)]
pub fn Ring<T: UIWidget + 'static>(width: u32, inner: T) -> RingWidget {
    RingWidget(Box::new(inner), width, None, false, 0, None)
}

pub struct RingWidget(
    // Inner
    Box<dyn UIWidget>,
    // Size
    u32,
    // Color
    Option<Box<dyn UIColor>>,
    // Inset
    bool,
    // Offset Width
    u32,
    // Offset Color
    Option<Box<dyn UIColor>>,
);

impl RingWidget {
    #[must_use]
    pub const fn inset(mut self) -> Self {
        self.3 = true;
        self
    }

    #[must_use]
    pub const fn offset_width(mut self, offset: u32) -> Self {
        self.4 = offset;
        self
    }

    #[must_use]
    pub fn offset_color<C: UIColor + 'static>(mut self, color: C) -> Self {
        self.5 = Some(Box::new(color));
        self
    }

    #[must_use]
    pub fn color<C: UIColor + 'static>(mut self, color: C) -> Self {
        self.2 = Some(Box::new(color));
        self
    }
}

impl Render for RingWidget {
    fn render(&self) -> Markup {
        self.render_with_class("")
    }
}

impl UIWidget for RingWidget {
    fn can_inherit(&self) -> bool {
        true
    }

    fn base_class(&self) -> Vec<String> {
        let class = match self.1 {
            0 => "ring-0",
            1 => "ring-1",
            2 => "ring-2",
            4 => "ring-4",
            8 => "ring-8",
            _ => &format!("ring-[{}px]", self.1),
        };

        let mut ret = vec![class.to_string()];

        if let Some(color) = &self.2 {
            ret.push(format!("ring-{}", color.color_class()));
        }

        if self.3 {
            ret.push("ring-inset".to_string());
        }

        ret.push(match self.4 {
            0 => "ring-offset-0".to_string(),
            1 => "ring-offset-1".to_string(),
            2 => "ring-offset-2".to_string(),
            4 => "ring-offset-4".to_string(),
            8 => "ring-offset-8".to_string(),
            _ => format!("ring-offset-[{}px]", self.4),
        });

        if let Some(color) = &self.5 {
            ret.push(format!("ring-offset-{}", color.color_class()));
        }

        ret
    }

    fn extended_class(&self) -> Vec<String> {
        let mut c = self.base_class();
        c.extend_from_slice(&self.0.extended_class());
        c
    }

    fn render_with_class(&self, class: &str) -> Markup {
        if self.0.as_ref().can_inherit() {
            self.0
                .as_ref()
                .render_with_class(&format!("{} {class}", self.base_class().join(" ")))
        } else {
            html! {
                div class=(format!("{} {class}", self.base_class().join(" "))) {
                    (self.0.as_ref())
                }
            }
        }
    }
}
